
/*
    a tap to intercept a shutdown

    - when a user presses the remote OFF button, this tap will
      intercept it, and generate an RKEY_Power event to all taps

    - an interested tap may shut down nicely, save state etc
    - a non-interested tap should ignore RKEY_Power (return RKEY_Power in param1)

    - this tap will then generate a 2nd RKEY_Power event
    - following this, the box will shutdown

    -how does it work?
    on pressing the remote 'OFF' button, a watchdog is enabled and the shutdown task, task_id = 0xef00 generated
    the OS will then schedule this task
    the shutdown task
        - disables watchdog
        - displays 'OFF' on the front panel
        - checks to see if a recording is taking place; if so pops up an 'are you sure?' dialog
        - stops recording
        - stops the playback
        - jumps into a dead loop (0x80108df4)
        - shutsdown (presumably due to the watchdog timing out)

    this tap hooks the shutdown event

    Once the 'OFF' key is pressed, the box will always shutdown after a few seconds unless the
    watchdog is disabled.  Once the watchdog is disabled, further presses of the 'OFF' button get ignored and
    do not cause a shutdown task.

    So, there is a bit of jiggery pokery needed to get this to work.

    1. press 'off'
    2. get 1st shutdown event hook, disable irq (watchdog), generate RKEY_Power event to all taps, return
    3. on 1st RKEY_Power event, generate another RKEY_Power event
    4. get 2nd shutdown event hook, return  (sometimes get this one before the RKEY_Power event activates)
    5. on 2nd RKEY_Power, renable irq (watchdog), generate shutdown task
    6. get 3rd shutdown event hook, run original task code; box will shutdown


    a user tap simply needs to monitor for the RKEY_Power event
    (and not expect to get any idle events after receiving it)
    (and ignore it if it receives it a second time)
    or just ignore it

    a user tap will also receive the hook event (event=0xf100). It must ignore it and MUST return it with param1 unchanged.

    ... note the hook functions can be used for lots of interesting things
    e.g.
        intercepting _any_ firmware functions
        grabbing the raw TS data
        etc, etc


    - this tap also demonstrates shutting down the box
    - call the function
        exTAP_Shutdown();           to shutdown the box


    v0.01 - initial release
    v0.02 -
    v0.03 - updated for exTap_v0.14
*/
#include <tap.h>
#include "exTap.h"
#include "tf5800.h"

#define ID_USERTAP 0xbdb00403
TAP_ID( ID_USERTAP );

TAP_PROGRAM_NAME("shutdown");
TAP_AUTHOR_NAME("bdb");
TAP_DESCRIPTION("A TAP to intercept a shutdown");
TAP_ETCINFO(__DATE__);

//----------------------------------------------------------------------------
char *tap_name;
type_menu main_menu = {
    0,
    0,
    {0},
    0,
    (char *)__tap_program_name__,
    {
        {5, "xxx",                      },
        {0, "Enable Shutdown Event",    },
        {1, "Shutdown",                 },
        {2, "Issue RKEY_Power",         },
        {3, "Exit Tap",                 },
    },
};

//----------------------------------------------------------------------------
//
//tsr_commander stuff
//
#ifdef _WIN32
#else
#include "TSRCommander.inc"

bool TSRCommanderExitTAP()
{
    close();
    return FALSE;
}

void TSRCommanderConfigDialog()
{
    ShowMenu(&main_menu);
}

int is_tsr_commander_running(void)
{
    int i;
    type_TAPProcess *TAPProcess;

    for (i=0; i<16; i++) {
        TAPProcess = (type_TAPProcess *)(fw_tap_table + (i*32));
        if (TAPProcess->firmwaregp) {
            if (TAPProcess->header) {
                if (TAPProcess->header->id == 0x807a10b2) {  //tsr_commander
                    return 1;
                }
                if (TAPProcess->header->id == 0x81001021) {  //tap_commander
                    return 1;
                }
            }
        }
    }
    return 0;
}
#endif

//----------------------------------------------------------------------------
int TAP_Main(void);
dword TAP_EventHandler(word event, dword param1, dword param2);
int menu_handler(type_menu *menu, dword param1);
int key_handler(dword param1);
int init(void);
int close(void);

//----------------------------------------------------------------------------
int shutdown_enabled = 0;
int shutdown_hook_id = -2;
int shutdown = 0;
int apply_shutdown_hook(void);
int remove_shutdown_hook(void);
dword shutdown_hook(dword param1, dword param2);

//----------------------------------------------------------------------------
int TAP_Main(void)
{
    int ret;

    tap_name =  (char *)__tap_program_name__;
    TAP_Print("\nopening %s", tap_name);
    ShowMessage("opening", tap_name, 100, 1);

    //determine support for different firmwares
    ret = exTAP_Init();
    if (ret) {
        return 0;
    }
    if ((fw_shutdown_irq_disable == 0) || (fw_shutdown_task == 0)) {
        return 0;
    }

    init();
    return 1;                 //0=application mode, 1=tsr mode
}

//----------------------------------------------------------------------------
dword TAP_EventHandler(word event, dword param1, dword param2)
{
    int ret;
    dword state, subState;

    switch (event) {
        case EVT_IDLE:
//            TAP_Print("\nEVT_IDLE, event=%x, param1=%x, param2=%x", event, param1, param2);

            //remove a message when it times out
            if (message_showing) {
                if (TAP_GetTick() > message_showing) {
                	TAP_Osd_Delete(message_rgn);    // release rgn-handle
                    message_showing = 0;
                    message_rgn = 0;
                }
            }
            break;


        case EVT_KEY:
///*
//master shutdown tap only
            //always process RKEY_Power, even if showing another menu
            if (param1 == RKEY_Power) {
                param1 = key_handler(param1);
            }
            else {
//*/
                TAP_GetState(&state, &subState);
                if (main_menu.visible) {
                    param1 = menu_handler(&main_menu, param1);
                }
                else {
                    if ((state == STATE_Normal) && (subState == SUBSTATE_Normal)) {
                        param1 = key_handler(param1);
                    }
                }
///*
            }
//*/
            break;

        case EVT_HOOK:
//            TAP_Print("\nEVT_HOOK, event=%x, param1=%x, param2=%x", event, param1, param2);
            if (param1 == shutdown_hook_id) {
                param1 = shutdown_hook(param1, param2);
            }
            break;

        default:
//TAP_Print("\nEVT_???, event=%x, param1=%x, param2=%x", event, param1, param2);
            break;
    }
    return param1;
}

int menu_handler(type_menu *menu, dword param1)
{
    int selection;
    int ret;

    switch (param1) {
        case RKEY_Ok:
            selection = TAP_Win_GetSelection(&menu->win);
            DeleteMenu(menu);
			switch (selection) {
				case 0:
				    if (shutdown_enabled) {
	    			    ret = remove_shutdown_hook();
                        if (ret == 0) strcpy(menu->item[1].str, "Enable Shutdown Event");
                        shutdown_enabled = 0;
				    }
				    else {
	    			    ret = apply_shutdown_hook();
                        if (ret == 0) {
                            strcpy(menu->item[1].str, "Disable Shutdown Event");
                            shutdown_enabled = 1;
                        }
				    }
				    return 0;
				case 1:
                    TAP_Print("\nexTAP_Shutdown()");
                    exTAP_Shutdown();
				    return 0;
				case 2:
                    TAP_Print("\nTAP_GenerateEvent(EVT_KEY, RKEY_Power, 0)");
                    TAP_GenerateEvent(EVT_KEY, RKEY_Power, 0);
				    return 0;
				case 3:
				    close();
				    TAP_Exit();
				    return 0;
			}
            break;
        case RKEY_Recall:
			DeleteMenu(menu);
            return param1;
        case RKEY_Exit:
			DeleteMenu(menu);
            return 0;
        default:
	        TAP_Win_Action(&menu->win, param1);
    		return 0;
    }
    return param1;
}

int key_handler(dword param1)
{
    static int seen_shutdown = 0;

    switch(param1) {
        case RKEY_Power:
//            TAP_Print("\nRKEY_Power");
///*
//only if master shutdown tap
            if (shutdown == 0) {
                //RKEY_Power event manually generated
                TAP_Print("\nRKEY_Power event manually generated");
            }
            else {
                if (!seen_shutdown) {
                    //
                    //save settings ...
                    //
                    TAP_Print("\nSaving Settings");

                    TAP_GenerateEvent(EVT_KEY, RKEY_Power, 0);
                    seen_shutdown = 1;
                }
                else {
                    TAP_Print("\nShutting Down");
//                    exTAP_CallFirmware(fw_shutdown_irq_enable, 0, 0,0,0,0,0,0,0,0,0,0);    // renable shutdown irq - not necessary since we will force a shutdown
                    exTAP_Shutdown();
                }
                shutdown++;
            }
//*/
/*
//if slave shutdown tap
//          TAP_Print("\nSaving Settings");
//
*/
            break;

        //only if tsrcommander is not running
        case RKEY_Recall:
            if (!is_tsr_commander_running()) {
                ShowMenu(&main_menu);
                return 0;
            }
            break;
    }
    return param1;
}

//----------------------------------------------------------------------------
int init(void)
{
    main_menu.visible = 0;
    TSRCommanderInit(0, TRUE);

    if (shutdown_enabled == 0) strcpy(main_menu.item[1].str, "Enable Shutdown Event");
    else                       strcpy(main_menu.item[1].str, "Disable Shutdown Event");

    return 0;
}

int close(void)
{
    TAP_Print("\nclosing %s", tap_name);
    ShowMessage("closing", tap_name, 100, 1);

    //all hooks MUST be removed before shutting down a tap
    remove_shutdown_hook();

    DeleteMenu(&main_menu);
    return 0;
}

//-----------------------------------------------------------------------------
//shutdown functions

type_hook shutdown_hook_patch[] = {
// sysid, firmware, enable, address,    length, original_data
    {456, 0x1209,   1,      0x00000000, 8,      {0x27, 0xbd, 0xff, 0xb8,  0xaf, 0xb4, 0x00, 0x38, }, },  //shutdown
    {0},
};

int apply_shutdown_hook(void)
{
    if (fw_shutdown_task == 0) {
        TAP_Print("\n%s(), Unsupported firmware: sysid:%d, version:%04x", __func__, fw_sysid, _appl_version);
        return -1;
    }

    shutdown_hook_patch[0].sysid = fw_sysid;
    shutdown_hook_patch[0].revision = _appl_version;
    shutdown_hook_patch[0].addr = (dword)fw_shutdown_task;

    TAP_Print("\nApplying Shutdown Hook");
    shutdown_hook_id = exTAP_ApplyHook(&shutdown_hook_patch[0]);
    return 0;
}

int remove_shutdown_hook(void)
{
    int ret;

    if (shutdown_hook_id >=0) {
        TAP_Print("\nRemoving Shutdown Hook");
        ret = exTAP_RemoveHook(shutdown_hook_id);
        if (ret == 0) shutdown_hook_id = -1;
    }
    return 0;
}

dword shutdown_hook(dword param1, dword param2)
{
    if (shutdown == 0) {
        exTAP_CallFirmware((void *)fw_shutdown_irq_disable, 0, 0,0,0,0,0,0,0,0,0,0);       //disable shutdown irq, prevents automatic shutdown on timeout
        shutdown ++;
        TAP_GenerateEvent(EVT_KEY, RKEY_Power, 0);
        return -1;                  //force rts; will not shutdown
    }
    //need this to ensure the RKEY_Power completes ...
    else if (shutdown == 1) {
        shutdown ++;
        return -1;                  //force rts; will not shutdown
    }
    else {
        shutdown = 0;
        return param1;              //allow continue; will shutdown
    }
    return param1;
}

//----------------------------------------------------------------------------
